
--! @file
--! @brief Hlavní adresový dekodér pro platformu Z102
--!

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

--! @brief
--! <b> Hlavní adresový dekodér pro platformu Z102 </b>
--!
--! Blok pro připojení procesoru (CPU) k FPGA pomocí paralelní 16-ti bitové sběrnice. 
entity MainAddressDecoder is
  generic(
    PllCount2Lock : integer; -- PLL-LOCKED minimum pulse length [PhyClk counts]
    UnitType      : string; -- unit type string
    UnitNumber    : integer; -- unit number
    UnitMV        : string -- unit mechanic version
  );
  port(
    --! Hodinový signál
    Clock           : in std_logic;
    --! Asynchronni reset
    Reset           : in std_logic;
    --! Globální vstup, informace o ukončení inicializačních procesů
    DeviceOk        : in std_logic;
    --! Zápisový povel (CPU), dolní BYTE - po navzorkování samplerem (např. STM32cpuSampler)
    WrL             : in std_logic;
    --! Zápisový povel (CPU), horní BYTE - po navzorkování samplerem (např. STM32cpuSampler)
    WrH             : in std_logic;
    --! Výběrové signály (Chip-select CPU)
    CsIn            : in std_logic_vector(3 downto 0);
    --! Adresa z CPU, po navzorkování samplerem (např. STM32cpuSampler)
    Address         : in std_logic_vector(23 downto 0);
    --! Data z CPU
    Din             : in std_logic_vector(15 downto 0);
    --! Data pro CPU
    Dout            : out std_logic_vector(15 downto 0);
    --! Zavěšení PLL - z výstupu hlavního fázového závěsu
    PllLocked       : in std_logic;
    --! @brief
    --! Hodinový signál pro kontrolu zavěšení PLL
    --! 
    --! @warning Nesmí být produktem kontrolovaného PLL !!
    --! 
    PllPhyClk       : in std_logic;
    --! Samostatný reset PLL, nepovinný
    PllReset        : in std_logic;
    --! Příznak prvního startu PLL (po reset PLL nebo konfiguraci FPGA), čeká na první správné zavěšení
    PllStartupWait  : out std_logic := '1';
    --! Výstup detektoru PLL
    PllReallyOk     : out std_logic := '0';
    --! Adresa z CPU, po navzorkování samplerem (např. STM32cpuSampler)
    Data        : std_logic_vector (31 downto 0);
    
    RackAddr        : in std_logic_vector(2 downto 0);
    RackParity      : in std_logic    
  );
end MainAddressDecoder;

--! @cond DOCUMENT_BODIES
--! 

--! @brief
--! <b> Hlavní adresový dekodér pro platformu Z102 - implementace </b>
--! 
--! Architektura dekodéru se skládá ze dvou procesů a kombinační logiky. 
--! @endcond
--! 
architecture ARCH of MainAddressDecoder is
--! @cond DOCUMENT_BODIES
--! 

  --! Začátek bloku vnitřního systémového prostoru dekodéru
  constant InsideBlockAddr : integer := 16#FE00#;
  --! Délka bloku vnitřního systémového prostoru dekodéru
  constant InsideBlockSize : integer := 16#0200#;
  --! Adresa číselného kódu typu desky (prefix, např. ZD)
  constant UnitTypeAddr    : integer := InsideBlockAddr + 16#10#;
  --! Adresa typu desky (číslo desky, např. 1234)
  constant UnitNumberAddr  : integer := InsideBlockAddr + 16#12#;
  --! Adresa číselného kódu typu desky (postfix, např. P1)
  constant UnitMVAddr      : integer := InsideBlockAddr + 16#14#;
  --! Adresa testovacích dat pro kontrolu funkčnosti FPGA při startu systému
  constant TestWordAddr    : integer := InsideBlockAddr + 16#40#;
  
  --! Adresa SN (sériové číslo) desky - WORD_0 (16 bit)
  constant SNAddrW0        : integer := InsideBlockAddr + 16#48#;
  --! Adresa SN (sériové číslo) desky - WORD_1 (16 bit)
  constant SNAddrW1        : integer := InsideBlockAddr + 16#4A#;
  --! Adresa SN (sériové číslo) desky - WORD_2 (16 bit)
  constant SNAddrW2        : integer := InsideBlockAddr + 16#4C#;
  --! Adresa SN (sériové číslo) desky - WORD_3 (16 bit)
  constant SNAddrW3        : integer := InsideBlockAddr + 16#4E#;
  
  --! Adresa hlavního systémového registru
  constant SysCtrlAddr     : integer := InsideBlockAddr + 16#80#;
                  -- Struktura hlavního systémového registru (SysCtrl), pozice bitů
  constant SysCtrl_RunLed_Bit        : integer := 15; --! Ovládání RUN LED
  constant SysCtrl_StopLed_Bit       : integer := 14; --! Ovládání STOP/ERROR LED
  constant SysCtrl_ParityError_Bit   : integer := 9; --! Chyba Parity (využíván ovladačem SSRAM - Ssram3portController)
  constant SysCtrl_PllError_Bit      : integer := 8; --! Chyba zavěšení PLL
  
  --! Adresa hlavního registru pro ovládání I/O
  constant IoCtrlAddr      : integer := InsideBlockAddr + 16#C0#;
                  -- Struktura hlavního registru pro ovládání I/O (IoCtrl), pozice bitů
  constant IoCtrl_Lock_Bit           : integer := 15; --! Zamykání I/O registrů pro konzistentní přístup z CPU
  
---------------------------------------

  --! @brief Výčtový typ pro časování přístupu CPU k FPGA
  type tFpgaTiming is (
    Timing_STD,
    Timing_12xM,
    Timing_64M,
    Timing_128M_WAIT_RAM_CS2,
    Timing_64M_WAIT_RAM_CS2
  );
  
  --! @brief Vstupní struktura portu SSRAM paměti (@ref Ssram3portController)
  type tSsramPortInput is record      
      Data        : std_logic_vector (31 downto 0); --! Vstupní data
      Address     : std_logic_vector (19 downto 0); --! Vstupní adresa
      BWEA        : std_logic ; -- Povel pro zápis BYTE A [1]
      BWEB        : std_logic ; -- Povel pro zápis BYTE B [1]
      BWEC        : std_logic ; -- Povel pro zápis BYTE C [1]
      BWED        : std_logic ; --! Povel pro zápis BYTE D [1]
      RDE         : std_logic ; --! Povel pro čtení [1]
  end record tSsramPortInput;
  
---------------------------------------

  --! Číselný kód typu desky (prefix) - info pro exekutivu CPU
  signal  iUnitType   : integer := 0;
  --! Číselný kód typu desky (postfix) - info pro exekutivu CPU
  signal  iUnitMV     : integer := 0;
  --! Chip-select "vnitřního prostoru" dekodéru
  signal  CsInside    : std_logic := '0';
  --! Chyba adresy vany (RackAddr)
  signal  RackAerr    : std_logic := '1';
  --! Data pro čtení z CPU
  signal  RdData      : std_logic_vector(15 downto 0) := (others => '0');
  --! Testovací data pro kontrolu funkčnosti FPGA při startu systému
  signal  TestData    : std_logic_vector(15 downto 0) := (others => '0');
  --! Hlavní systémový registr
  signal  SysCtrlData : std_logic_vector(15 downto 0) := (others => '0');
  --! Hlavní registr pro ovládání I/O
  signal  IoCtrlData  : std_logic_vector(15 downto 0) := (others => '0');
  --! Příznak správné funkce PLL
  signal  PllOk       : std_logic := '0';
  --! Chyba PLL (zavěšení)
  signal  PllError    : std_logic := '0';
  --! Mazání chyby zavěšení PLL
  signal  PllErrorClr : std_logic := '1';
 
  signal  FpgaTiming    : tFpgaTiming := Timing_12xM; -- advanced FPGA timing setup
  
---------------------------------------

  --! @brief Konverze typů : integer => std_logic_vector
  --! @param X Vstupní hodnota
  --! @param L Požadovaná šířka výstupního slova v bitech
  --! @return Převedená hodnota
  --!
  --! Funkce převádí typ integer na typ std_logic_vector.
  function ToStdLogicVector(X,L:integer) return std_logic_vector is
    begin
      if X < 0 then
        return std_logic_vector(to_signed(X,L));
      else
        return std_logic_vector(to_unsigned(X,L));
      end if;
    end ToStdLogicVector;

  --! @brief Konverze typů : integer => std_logic_vector(16)
  --! @param X Vstupní hodnota
  --! @return Převedená hodnota
  --!
  --! Funkce převádí typ integer na 16-ti bitové slovo typu std_logic_vector.
  function ToVector16(constant X:integer) return std_logic_vector is
    begin
      return ToStdLogicVector(X, 16);
    end ToVector16;

  --! @brief Konverze typů : integer => std_logic_vector(24)
  --! @param X Vstupní hodnota
  --! @return Převedená hodnota
  --!
  --! Funkce převádí typ integer na 24-ti bitové slovo typu std_logic_vector.
  function ToVector24(constant X:integer) return std_logic_vector is
    begin
      return ToStdLogicVector(X, 24);
    end ToVector24;

  --! Generátor/detektor parity
  component ParityCheck
    generic(
      Size      : natural
    );
    port(
      Data      : in std_logic_vector(Size-1 downto 0);
      ParityBit : in std_logic;
      Q         : out std_logic
    );
  end component ParityCheck;

---------------------------------------
--! @endcond
--! 
  begin

    -- All components used in architecture, always for DOXYGEN :

    --! Vyhonoceni parity RACK adresy
    RackPar : ParityCheck
      generic map(Size => 3)
      port map(Data => RackAddr, ParityBit => not(RackParity), Q => RackAerr);
              
--! @cond DOCUMENT_BODIES
--! 

    iUnitType <= character'pos(UnitType(1)) * 256 + character'pos(UnitType(2)); -- UnitType compute
    iUnitMV <= character'pos(UnitMV(1)) * 256 + character'pos(UnitMV(2)); -- UnitMV compute

    -- TestData conditions
    TestData <= ToVector16(16#1234#) when DeviceOk = '1' and PllOk = '1' else (others => '0');

    -- system control & status data
    SysCtrlData(SysCtrl_PllError_Bit) <= PllError; -- PLL error flag
    SysCtrlData(2 downto 0) <= ToStdLogicVector(FpgaTiming'pos(FpgaTiming), 3);
    

    CsInside <= '1' when CsIn(0) = '1' and (Address >= ToVector24(InsideBlockAddr)) and -- main "inside" dekoder space chip-select
              (Address < ToVector24(InsideBlockAddr+InsideBlockSize))
        else '0';
        	
    Dout <= RdData; -- data output
    
    with Address(15 downto 0) select -- RD address decoder, "inside" decoder data space
      RdData <= ToVector16(iUnitType)                     when ToVector16(UnitTypeAddr),
                ToVector16(UnitNumber)                    when ToVector16(UnitNumberAddr),
                ToVector16(iUnitMV)                       when ToVector16(UnitMVAddr),
                TestData                                  when ToVector16(TestWordAddr),
                SysCtrlData                               when ToVector16(SysCtrlAddr),
                IoCtrlData                                when ToVector16(IoCtrlAddr),
                ToVector16(0)                             when others;

    --! @brief Zápis dat z CPU
    --!
    --! Proces zajišťuje : 
    --!
    --! Zápis dat do vnitřních registrů adresového dekodéru : 
    --! @vhdlflow Vývojový diagram procesu DataWrite v architektuře MainAddressDecoder
    DataWrite : process(Reset, Clock) -- write data to registers
      variable SNAddrCs : std_logic_vector (7 downto 0) := (others => '0'); -- registered for improve timing ...
    begin
      if Reset = '1' then -- async. resets (by MainReset)
        IoCtrlData(IoCtrl_Lock_Bit) <= '0';
        SysCtrlData(SysCtrl_RunLed_Bit) <= '0';
        SysCtrlData(SysCtrl_StopLed_Bit) <= '0';
        SNAddrCs := (others => '0');
        PllErrorClr <= '1'; -- clear after reset
        SysCtrlData(SysCtrl_ParityError_Bit) <= '0';
      elsif Clock'event and Clock = '1' then -- main clock active edge
        PllErrorClr <= '0'; -- prepare NO CLEAR ...
        
        if CsInside = '1' then -- is it MY internal space ?
          if WrH = '1' then -- high byte WR
            case Address(15 downto 0) is -- by address
              when ToVector16(IoCtrlAddr)   =>  IoCtrlData(IoCtrl_Lock_Bit) <= Din(IoCtrl_Lock_Bit); -- I/O control register
              when ToVector16(SysCtrlAddr)  =>  SysCtrlData(SysCtrl_RunLed_Bit) <= Din(SysCtrl_RunLed_Bit); -- system control register
                                                SysCtrlData(SysCtrl_StopLed_Bit) <= Din(SysCtrl_StopLed_Bit);
                                                -- CLEAR parity error
                                                SysCtrlData(SysCtrl_ParityError_Bit) <= SysCtrlData(SysCtrl_ParityError_Bit) and
                                                                                        Din(SysCtrl_ParityError_Bit);
                                                PllErrorClr <= PllError and not(Din(SysCtrl_PllError_Bit)); -- CLEAR pulse for PLLerror
              when ToVector16(SNAddrW0)     => SNAddrCs(1) := '1'; -- SN registers
              when ToVector16(SNAddrW1)     => SNAddrCs(3) := '1';
              when ToVector16(SNAddrW2)     => SNAddrCs(5) := '1';
              when ToVector16(SNAddrW3)     => SNAddrCs(7) := '1';
              when others => null;
            end case;
          end if;
          if WrL = '1' then -- low byte WR
            case Address(15 downto 0) is -- by address
              when ToVector16(SNAddrW0)     => SNAddrCs(0) := '1'; -- SN registers
              when ToVector16(SNAddrW1)     => SNAddrCs(2) := '1';
              when ToVector16(SNAddrW2)     => SNAddrCs(4) := '1';
              when ToVector16(SNAddrW3)     => SNAddrCs(6) := '1';
              when others => null;
            end case;
          end if;
        end if;
      end if;
      
    end process DataWrite;

    --! @brief Kontrola stability fázového závěsu (PLL)
    --!
    --! Kontrola zavěšení PLL
    --! @vhdlflow Vývojový diagram procesu PllCheck v architektuře MainAddressDecoder
    PllCheck : process(Reset, PllReset, PllErrorClr, PllPhyClk, PllLocked)
      variable  PllCounter : natural range 0 to PllCount2Lock := PllCount2Lock;
      variable  PllStartup : boolean := true; -- wait for first PLL lock
    begin
      -- PLL ERROR detector
      if Reset = '1' or PllErrorClr = '1' then -- clear conditions
        PllError <= '0'; -- reset PLL error flag
      elsif falling_edge(PllLocked) then -- error edge
        if not(PllStartup) then -- after first PLL lock
          PllError <= '1'; -- set PLL error flag
        end if;
      end if;
      
      -- PLL OK detector
if PllCount2Lock > 0 then -- PLL detector used (conditional compilation)
      if PllReset = '1' then -- PLL detector special restart
        PllStartup := true;
        PllCounter := PllCount2Lock;
        PllError <= '0';
      elsif rising_edge(PllPhyClk) then
        PllOk <= '0'; -- prepare false output
        if (PllLocked) = '0' then -- PLL error
          PllCounter := PllCount2Lock; -- restart counter
        else
          if PllCounter > 0 then -- wait for realy OK
            PllCounter := PllCounter - 1; -- decrement
          else
            PllOk <= '1'; -- output OK, counter == 0
            PllStartup := false; -- mark first PLL lock
          end if;
        end if;
      end if;
else
      -- PLL detector not used - direct PllLocked signal
      PllOk <= PllLocked;
end if;

      if PllStartup then -- STARTUP wait output
        PllStartupWait <= '1';
      else
        PllStartupWait <= '0';
      end if;

    end process PllCheck;
    
    PllReallyOk <= PllOk; -- set output signal PLL OK
      
--! @endcond
--! 
end ARCH;

